import React, { useState } from "react";
import { observer } from "mobx-react";
import { Button, Space, Tag } from "antd";
import {
  CredentialRecordDto,
  CredentialRecordDtoState,
} from "@dashboard/engine-api";
import format from "date-fns/format";
import api from "@dashboard/engine-api";
import modalStore from "@dashboard/store/modalStore";
import RecordItem from "../RecordItem";
import s from "./styles.module.scss";
import displayError from "@dashboard/utils/displayError";

interface CredentialItemProps {
  item: CredentialRecordDto;
  onOpen?: () => void;
  onRemove?: () => void;
  onAccept?: () => void;
  onDecline?: () => void;
  statusChanging?: boolean;
}

const CredentialItem = observer(
  ({
    item,
    onOpen,
    onRemove,
    onAccept,
    onDecline,
    statusChanging,
  }: CredentialItemProps) => {
    const [loadingFormatData, setLoadingFormatData] = useState(false);
    const [formatData, setFormatData] = useState<unknown>(null);

    const openFormatData = async () => {
      if (!formatData) {
        setLoadingFormatData(true);
        try {
          const formatData = await api.getCredentialFormatDataById(item.id!);
          setFormatData(formatData);
          setLoadingFormatData(false);
          modalStore.openViewJsonDialog({ data: formatData });
          return;
        } catch (e: unknown) {
          displayError(e);
          setLoadingFormatData(false);
          return;
        }
      }

      modalStore.openViewJsonDialog({ data: formatData });
    };

    return (
      <RecordItem rawData={item}>
        <div className={s.prop}>
          Status: <Tag>{item.state}</Tag>
        </div>
        <div className={s.prop}>
          Attbiutes:{" "}
          {item.attributes?.map((p) => (
            <Tag color="green" key={p.name}>
              {p.name}: {p.value}
            </Tag>
          ))}
        </div>
        <div className={s.createdAt}>
          {item.createdAt && format(item.createdAt, "dd.MM.yyyy HH:mm:ss")}
        </div>
        <Space direction="horizontal">
          {onOpen && <Button onClick={onOpen}>Go to credential</Button>}
          <Button
            type="primary"
            disabled={loadingFormatData}
            loading={loadingFormatData}
            onClick={openFormatData}
          >
            View Format Data
          </Button>
          {item.state === CredentialRecordDtoState.OfferReceived &&
            onAccept && (
              <Button
                disabled={statusChanging}
                onClick={onAccept}
                type="primary"
              >
                Accept
              </Button>
            )}
          {item.state === CredentialRecordDtoState.OfferReceived &&
            onDecline && (
              <Button disabled={statusChanging} onClick={onDecline} danger>
                Decline
              </Button>
            )}
          {onRemove && (
            <Button disabled={statusChanging} danger onClick={onRemove}>
              Delete
            </Button>
          )}
        </Space>
      </RecordItem>
    );
  },
);

export default CredentialItem;
