import { makeAutoObservable, runInAction } from "mobx";
import api, { ProofRecordDto } from "@dashboard/engine-api";
import { toast } from "react-toastify";
import displayError from "@dashboard/utils/displayError";

class ProofListStore {
  public loading = false;
  public items: ProofRecordDto[] = [];

  public statusChangeId: string | null = null;

  constructor() {
    makeAutoObservable(this);
  }

  public async loadAll(connectionId: string | null) {
    runInAction(() => (this.loading = true));
    try {
      const items = await api.fetchProofs(undefined, connectionId || undefined);
      runInAction(() => {
        this.items = items;
        this.loading = false;
      });
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => (this.loading = false));
    }
  }

  public async accept(proofRecordId: string) {
    runInAction(() => {
      this.statusChangeId = proofRecordId;
    });
    try {
      const credRecord = await api.acceptProof({
        proofId: proofRecordId,
        proofUrl: "",
      });
      runInAction(() => {
        const index = this.items.findIndex((p) => p.id === credRecord.id);
        if (index >= 0) {
          this.items[index] = credRecord;
        }
        this.statusChangeId = null;
      });
      toast.success("Proof request accepted and credential data is shared");
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.loading = false;
        this.statusChangeId = null;
      });
    }
  }

  public async decline(proofRecordId: string) {
    runInAction(() => {
      this.statusChangeId = proofRecordId;
    });
    try {
      const credRecord = await api.declineProofRequest(proofRecordId);
      runInAction(() => {
        const index = this.items.findIndex((p) => p.id === credRecord.id);
        if (index >= 0) {
          this.items[index] = credRecord;
        }
        this.statusChangeId = null;
      });
      toast.success("Proof request declined");
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.loading = false;
        this.statusChangeId = null;
      });
    }
  }

  public async remove(proofRecordId: string) {
    runInAction(() => {
      this.statusChangeId = proofRecordId;
    });
    try {
      await api.deleteProofById(proofRecordId);
      runInAction(() => {
        const index = this.items.findIndex((p) => p.id === proofRecordId);
        if (index >= 0) {
          this.items.splice(index, 1);
        }
        this.statusChangeId = null;
      });
      toast.success("proof request deleted successfully");
    } catch (e: unknown) {
      displayError(e);
      runInAction(() => {
        this.loading = false;
        this.statusChangeId = null;
      });
    }
  }
}

export type { ProofListStore };

export default ProofListStore;
