import {
  Body,
  Controller,
  Get,
  Param,
  Post,
  Delete,
  UseFilters,
  Query,
} from "@nestjs/common";
import {
  AcceptProofDto,
  GatewayAcceptedResponseDto,
  RequestProofDto,
  ProofFilterDto,
  PROOF_ACCEPT,
  PROOF_DECLINE,
  PROOF_GET,
  PROOF_REQUEST,
  PROOF_LIST,
  PROOF_DELETE,
  IdReqDto,
} from "@ocm-engine/dtos";
import { AllExceptionsHandler } from "../exception.handler";
import { ProofManagerClient } from "@ocm-engine/clients";
import { ApiBadRequestResponse, ApiOperation } from "@nestjs/swagger";
import IncludeOcmDtos from "../decorators/IncludeOcmDtos";
import OcmGatewayResponse from "../decorators/OcmGatewayResponse";
import OcmInternalServerErrorResponse from "../decorators/OcmInternalServerErrorResponse";

@UseFilters(AllExceptionsHandler)
@Controller("v1")
@IncludeOcmDtos()
export class ProofController {
  constructor(private readonly pmClient: ProofManagerClient) {}

  @Get("/proofs")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("proof manager")
  @ApiOperation({
    summary: "List received unaccepted proofs",
    description:
      "This method list proofs based on provided filters. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Proof"],
  })
  async fetchProofs(
    @Query() proofFilterDto: ProofFilterDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.pmClient.sendPayload({
      pattern: "proofs",
      payload: {
        source: "/proofs",
        data: proofFilterDto,
        type: PROOF_LIST,
      },
    });
  }

  @Get("/proofs/:proof_record_id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("proof manager")
  @ApiOperation({
    summary: "Get a single proof record by providing proof record id.",
    description:
      "Method get proof by id. Status - request-receive. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Proof"],
  })
  async getProofById(
    @Param("proof_record_id") proofRecordId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = proofRecordId;

    return this.pmClient.sendPayload<IdReqDto>({
      pattern: "proofs",
      payload: {
        source: "/credentials/proofs/:proof_record_id",
        data,
        type: PROOF_GET,
      },
    });
  }

  @Post("/proofs/request")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("proof manager")
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "attributes must contain at least 1 elements",
                "attributes must be an array",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Request proof for credential",
    description:
      "Method will request proof. If connection id is not passed, the proof will be OOB. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Proof"],
  })
  async requestProof(
    @Body() requestProofDto: RequestProofDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.pmClient.sendPayload<RequestProofDto>({
      pattern: "proofs",
      payload: {
        source: "/proofs/request",
        data: requestProofDto,
        type: PROOF_REQUEST,
      },
    });
  }

  @Post(`/proofs/accept`)
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("proof manager")
  @ApiOperation({
    summary: "Accept credential proof",
    description:
      "Method accept credential proof. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Proof"],
  })
  async acceptProof(
    @Body() acceptProofRequestDto: AcceptProofDto,
  ): Promise<GatewayAcceptedResponseDto> {
    return this.pmClient.sendPayload<AcceptProofDto>({
      pattern: "proofs",
      payload: {
        source: "/proofs/accept",
        data: acceptProofRequestDto,
        type: PROOF_ACCEPT,
      },
    });
  }

  @Post("/proofs/:proof_record_id/decline")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("proof manager")
  @ApiOperation({
    summary: "Decline a proof request.",
    description:
      "Method to decline a proof request by id. Status - request-receive. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Proof"],
  })
  async declineProofRequest(
    @Param("proof_record_id") proofRecordId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = proofRecordId;

    return this.pmClient.sendPayload<IdReqDto>({
      pattern: "proofs",
      payload: {
        source: "/proofs/:proof_record_id/decline",
        data,
        type: PROOF_DECLINE,
      },
    });
  }

  @Delete("/proofs/:proof_record_id")
  @OcmGatewayResponse()
  @OcmInternalServerErrorResponse("proof manager")
  @ApiOperation({
    summary: "Delete a proof record by its ID",
    description:
      "This method deletes a specific proof based on the provided ID. The ID of the response will be matched when you receive an event from the websocket",
    tags: ["Credentials Proof"],
  })
  async deleteProofById(
    @Param("proof_record_id") proofRecordId: string,
  ): Promise<GatewayAcceptedResponseDto> {
    const data = new IdReqDto();
    data.id = proofRecordId;

    return this.pmClient.sendPayload<IdReqDto>({
      pattern: "proofs",
      payload: {
        source: "/proofs/:proof_record_id",
        data,
        type: PROOF_DELETE,
      },
    });
  }
}
