import {
  Body,
  Query,
  Controller,
  Delete,
  Get,
  Param,
  Post,
  UseFilters,
  UseGuards,
} from "@nestjs/common";

import { AgentService } from "../askar/agent.service";
import {
  CreateCredentialDefinitionRequestDto,
  OfferCredentialRequestDto,
  RequestProofDto,
  CreateSchemaRequestDto,
  MakeBasicMessageRequestDto,
  AcceptProofDto,
  IdReqDto,
  CredentialFilterDto,
  MessageFilterDto,
  ProofFilterDto,
  AcceptCredentialDto,
  CreateInvitationRequestDto,
  InvitationFilterDto,
  AcceptInvitationRequestDto,
  OfferJsonCredentialRequests,
  SignJsonCredentialRequests,
  DidRecordDto,
} from "@ocm-engine/dtos";
import { AllExceptionsHandler } from "./exception.handler";
import { DidResolutionResult } from "@credo-ts/core";
import { AuthGuard } from "./auth/auth.guard";

@UseFilters(AllExceptionsHandler)
@Controller("v1")
@UseGuards(AuthGuard)
export class RestController {
  constructor(private readonly agentService: AgentService) {}

  @Get("/invitations")
  async fetchInvitations(@Query() filter: InvitationFilterDto) {
    return this.agentService.fetchInvitations(filter);
  }

  @Post("/invitations")
  createInvitation(
    @Body() createInvitationRequestDto: CreateInvitationRequestDto,
  ) {
    return this.agentService.createInvitation(createInvitationRequestDto);
  }

  @Get("/invitations/:id")
  getInvitationById(@Param("id") id: string) {
    return this.agentService.getInvitationById(id);
  }

  @Delete("/invitations/:id")
  deleteInvitationById(@Param("id") id: string) {
    return this.agentService.deleteInvitationById(id);
  }

  @Post("/invitations/accept")
  async acceptInvitation(
    @Body() createInvitationDto: AcceptInvitationRequestDto,
  ) {
    const url =
      createInvitationDto.invitationUrl ||
      createInvitationDto.shortInvitationUrl;

    return this.agentService.acceptInvitation(url);
  }

  @Get("/connections")
  async fetchConnections() {
    return this.agentService.fetchConnections();
  }

  @Get("/connections/:id")
  async getConnectionById(@Param("id") id: string) {
    return this.agentService.getConnectionById(id);
  }

  @Delete("/connections/:id")
  async deleteConnectionById(@Param("id") id: string) {
    return this.agentService.deleteConnectionById(id);
  }

  @Get("/connections/oob/:id")
  async getConnectionByOobId(@Param("id") id: string) {
    return this.agentService.getConnectionByOobId(id);
  }

  @Post("/connections/ping/:id")
  async pingConnection(@Param("id") id: string) {
    return this.agentService.trustPingToConnection(id);
  }

  @Post("/schemas")
  async createSchema(@Body() schemaDto: CreateSchemaRequestDto) {
    return this.agentService.createSchema(schemaDto);
  }

  @Post("/schemas/get-by-id")
  async getSchemaById(@Body() dto: IdReqDto) {
    return this.agentService.getSchemaById(dto.id);
  }

  @Get("/schemas")
  async fetchSchemas() {
    return this.agentService.fetchSchemas();
  }

  @Get("/definitions")
  async fetchCredentialDefinitions() {
    return this.agentService.fetchCredentialDefinitions();
  }

  @Post("/definitions/get-by-id")
  async getCredentialDefinitionById(@Body() dto: IdReqDto) {
    return this.agentService.getCredentialDefinitionById(dto.id);
  }

  @Post("/definitions")
  async createCredentialDefinition(
    @Body() credentialDefinitionDto: CreateCredentialDefinitionRequestDto,
  ) {
    return this.agentService.createCredentialDefinition(
      credentialDefinitionDto,
    );
  }

  @Post("/credentials/offers")
  async offerCredential(@Body() dto: OfferCredentialRequestDto) {
    return this.agentService.offerCredential(dto);
  }

  @Post("/credentials/jsonld/offers")
  async offerJsonLdCredential(@Body() data: OfferJsonCredentialRequests) {
    return this.agentService.offerJsonLdCredential(data.connectionId, data.doc);
  }

  @Post("/jsonld/sign")
  async signJsonLdCredential(@Body() data: SignJsonCredentialRequests) {
    return this.agentService.signJsonLdCredential(data.doc);
  }

  @Post("/jsonld/credentials/:cred_id/prepare-verifiable-presentation")
  async signJsonLdPresentationByCredId(@Param("cred_id") credentialId: string) {
    return this.agentService.prepareVerifiablePresentationByJsonLdCredId(
      credentialId,
    );
  }

  @Get("/credentials")
  async fetchCredentials(@Query() credentialFilterDto: CredentialFilterDto) {
    return this.agentService.fetchCredentials(credentialFilterDto);
  }

  @Get("/credentials/:id")
  async getCredentialById(@Param("id") credentialId: string) {
    return this.agentService.getCredentialById(credentialId);
  }

  @Get("/credentials/:id/format-data")
  async getCredentialFormatDataById(@Param("id") credentialId: string) {
    return this.agentService.getCredentialFormatDataById(credentialId);
  }

  @Post("/credentials/offers/accept")
  async acceptCredential(@Body() dto: AcceptCredentialDto) {
    return this.agentService.acceptCredential(dto);
  }

  @Post("/credentials/offers/:credential_record_id/decline")
  async declineCredential(
    @Param("credential_record_id") credentialRecordId: string,
  ) {
    return this.agentService.declineCredential(credentialRecordId);
  }

  @Delete("/credentials/:id")
  async deleteCredentialById(@Param("id") credentialId: string) {
    return this.agentService.deleteCredentialById(credentialId);
  }

  @Post("/messages")
  async sendMessage(@Body() message: MakeBasicMessageRequestDto) {
    return this.agentService.sendMessage(message);
  }

  @Get("/messages")
  async fetchBasicMessages(@Query() filter: MessageFilterDto) {
    return this.agentService.fetchBasicMessages(filter);
  }

  @Delete("/messages/:id")
  async deleteBasicMessage(@Param("id") messageId: string) {
    return this.agentService.deleteMessageById(messageId);
  }

  @Get("/proofs")
  async fetchProofs(@Query() proofFilterDto: ProofFilterDto) {
    return this.agentService.fetchProofs(proofFilterDto);
  }

  @Get("/proofs/:proof_record_id")
  async getProofById(@Param("proof_record_id") proofRecordId: string) {
    return this.agentService.getProofById(proofRecordId);
  }

  @Get("/proofs/:proof_record_id/format-data")
  async getProofFormatDataById(
    @Param("proof_record_id") proofRecordId: string,
  ) {
    return this.agentService.getProofFormatDataById(proofRecordId);
  }

  @Post("/proofs/:proof_record_id/acceptance-wait")
  async proofAcceptanceWait(@Param("proof_record_id") proofRecordId: string) {
    return this.agentService.proofAcceptanceWait(proofRecordId);
  }

  @Post("/proofs/request")
  async requestProof(@Body() requestProofDto: RequestProofDto) {
    return this.agentService.requestProof(requestProofDto);
  }

  @Post(`/proofs/accept`)
  async acceptProof(@Body() acceptProofRequestDto: AcceptProofDto) {
    return this.agentService.acceptProof(acceptProofRequestDto);
  }

  @Post("/proofs/:proof_record_id/decline")
  async declineProofRequest(@Param("proof_record_id") proofRecordId: string) {
    return this.agentService.declineProofRequest(proofRecordId);
  }

  @Delete("/proofs/:proof_record_id")
  async deleteProofById(@Param("proof_record_id") proofRecordId: string) {
    return this.agentService.deleteProofById(proofRecordId);
  }

  @Get("/created-dids")
  async getCreatedDids(): Promise<DidRecordDto[]> {
    return this.agentService.getCreatedDids();
  }

  @Post("/resolve-did")
  async resolveDid(@Body() dto: IdReqDto): Promise<DidResolutionResult> {
    return this.agentService.resolve(dto.id);
  }
}
