import { Injectable, Logger } from "@nestjs/common";
import { AskarService } from "./askar.service";
import {
  ConnectionRecordDto,
  AcceptProofDto,
  CreateCredentialDefinitionRequestDto,
  CreddefRecordDto,
  CreateInvitationResponseDto,
  CreateSchemaRequestDto,
  SchemaRecordDto,
  CredentialNotCreatedError,
  OfferCredentialRequestDto,
  CredentialRecordDto,
  RequestProofDto,
  ProofRecordDto,
  MakeBasicMessageRequestDto,
  MessageRecordDto,
  SchemaNotCreatedError,
  CredentialFilterDto,
  ProofFilterDto,
  MessageFilterDto,
  RequestProofResponseDto,
  CredentialOfferResponseDto,
  EntityNotFoundError,
  AcceptCredentialDto,
  CredentialFormatDataDto,
  ProofFormatDataDto,
  CreateInvitationRequestDto,
  InvitationFilterDto,
  DidRecordDto,
  OcmError,
  W3cCredentialDto,
  BaseRecordDto,
  W3cJsonLdVerifiableCredentialDto,
  W3cJsonLdVerifiablePresentationDto,
} from "@ocm-engine/dtos";
import {
  AutoAcceptCredential,
  BasicMessageRecord,
  BasicMessageRole,
  ClaimFormat,
  ConnectionRecord,
  CredentialExchangeRecord,
  JsonTransformer,
  CredentialState,
  ProofState,
  Query,
  ProofExchangeRecord,
  OutOfBandRecord,
  W3cCredential,
  W3cCredentialService,
  DidRecord,
  JsonCredential,
  W3cJsonLdVerifiableCredential,
} from "@credo-ts/core";
import { AnonCredsRequestedAttribute } from "@credo-ts/anoncreds";
import { uuid } from "@credo-ts/core/build/utils/uuid";
import {
  waitForCredentialExchangeRecordSubject,
  waitForProofExchangeRecordSubject,
} from "../agent.utils";

@Injectable()
export class AgentService {
  private readonly logger = new Logger(AgentService.name);
  constructor(private readonly askar: AskarService) {}

  createInvitation = async (
    createInvitationRequestDto?: CreateInvitationRequestDto,
  ) => {
    const outOfBoundRecord = await this.askar.agent.oob.createInvitation(
      createInvitationRequestDto,
    );

    const response = new CreateInvitationResponseDto();

    let longUrl = outOfBoundRecord.outOfBandInvitation.toUrl({
      domain: this.askar.agentConfig.agentPeerAddress,
    });

    if (this.askar.agentConfig.agentOobUrl) {
      longUrl = longUrl.replace(
        this.askar.agentConfig.agentPeerAddress,
        this.askar.agentConfig.agentOobUrl,
      );
    }

    //TODO: should we replace the short url with agentOobUrl if we do this we should have a redirect in ingress
    response.shortInvitationUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBoundRecord.outOfBandInvitation.id}`;
    response.outOfBandId = outOfBoundRecord.id;
    response.createdAt = outOfBoundRecord.createdAt;
    response.updatedAt = outOfBoundRecord.updatedAt;
    response.role = outOfBoundRecord.role;
    response.state = outOfBoundRecord.state;
    response.invitationUrl = longUrl;

    return response;
  };

  acceptInvitation = async (
    invitationUrl: string,
  ): Promise<ConnectionRecordDto> => {
    const { connectionRecord } =
      await this.askar.agent.oob.receiveInvitationFromUrl(invitationUrl);

    if (typeof connectionRecord === "undefined") {
      throw new EntityNotFoundError();
    }

    const response = new ConnectionRecordDto();
    response.connectionName = connectionRecord.theirLabel;
    response.state = connectionRecord.state;
    response.id = connectionRecord.id;
    response.did = connectionRecord.did;
    response.invitationDid = connectionRecord.invitationDid;
    response.outOfBandId = connectionRecord.outOfBandId;
    response.createdAt = connectionRecord.createdAt;
    response.updatedAt = connectionRecord.updatedAt;
    response.imageUrl = connectionRecord.imageUrl;

    return response;
  };

  deleteInvitationById = async (id: string) => {
    return this.askar.agent.oob.deleteById(id);
  };

  fetchInvitations = async (filter: InvitationFilterDto) => {
    const query: Query<OutOfBandRecord>[] = [];

    if (filter.states) {
      const stateQuery: Query<OutOfBandRecord> = {
        $or: filter.states.map((state) => ({ state })),
      };
      query.push(stateQuery);
    }

    if (filter.roles) {
      const roleQuery: Query<OutOfBandRecord> = {
        $or: filter.roles.map((role) => ({ role })),
      };

      query.push(roleQuery);
    }

    let invitations = await this.askar.agent.oob.findAllByQuery({
      $and: query,
    });
    invitations = invitations.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const invitationsResponse = invitations.map((invitation) => {
      const response = new CreateInvitationResponseDto();
      response.invitationUrl = invitation.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });
      response.shortInvitationUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${invitation.outOfBandInvitation.id}`;
      response.outOfBandId = invitation.id;
      response.createdAt = invitation.createdAt;
      response.updatedAt = invitation.updatedAt;
      response.role = invitation.role;
      response.state = invitation.state;
      return response;
    });

    return invitationsResponse;
  };

  getInvitationById = async (oobId: string) => {
    const invitation = await this.askar.agent.oob.getById(oobId);

    const response = new CreateInvitationResponseDto();
    response.invitationUrl = invitation.outOfBandInvitation.toUrl({
      domain: this.askar.agentConfig.agentPeerAddress,
    });
    response.shortInvitationUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${invitation.outOfBandInvitation.id}`;
    response.outOfBandId = invitation.id;
    response.createdAt = invitation.createdAt;
    response.updatedAt = invitation.updatedAt;
    response.role = invitation.role;
    response.state = invitation.state;
    return response;
  };

  async fetchConnections(): Promise<ConnectionRecordDto[]> {
    //TODO: no ordering in findAllByQuery
    const agentResponse = (await this.askar.agent.connections.getAll()).sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const connectionArray = agentResponse.map((singleConnectionRes) => {
      const connectionResponse = new ConnectionRecordDto();
      connectionResponse.id = singleConnectionRes.id;
      connectionResponse.state = singleConnectionRes.state;
      connectionResponse.connectionName = singleConnectionRes.theirLabel;
      connectionResponse.alias = singleConnectionRes.alias;
      connectionResponse.did = singleConnectionRes.did;
      connectionResponse.invitationDid = singleConnectionRes.invitationDid;
      connectionResponse.outOfBandId = singleConnectionRes.outOfBandId;
      connectionResponse.createdAt = singleConnectionRes.createdAt;
      connectionResponse.updatedAt = singleConnectionRes.updatedAt;
      connectionResponse.imageUrl = singleConnectionRes.imageUrl;

      return connectionResponse;
    });

    return connectionArray;
  }

  getConnectionByOobId = async (oobId: string) => {
    const connectionRecords =
      await this.askar.agent.connections.findAllByOutOfBandId(oobId);

    const connectionArray = connectionRecords.map((singleConnectionRes) => {
      const connectionResponse = new ConnectionRecordDto();
      connectionResponse.id = singleConnectionRes.id;
      connectionResponse.state = singleConnectionRes.state;
      connectionResponse.connectionName = singleConnectionRes.theirLabel;
      connectionResponse.alias = singleConnectionRes.alias;
      connectionResponse.did = singleConnectionRes.did;
      connectionResponse.invitationDid = singleConnectionRes.invitationDid;
      connectionResponse.outOfBandId = singleConnectionRes.outOfBandId;
      connectionResponse.createdAt = singleConnectionRes.createdAt;
      connectionResponse.updatedAt = singleConnectionRes.updatedAt;
      connectionResponse.imageUrl = singleConnectionRes.imageUrl;

      return connectionResponse;
    });

    return connectionArray;
  };

  getConnectionById = async (id: string): Promise<ConnectionRecordDto> => {
    const agentResponse = await this.askar.agent.connections.findById(id);

    if (!agentResponse) {
      throw new EntityNotFoundError();
    }

    const connectionResponse = new ConnectionRecordDto();
    connectionResponse.id = agentResponse.id;
    connectionResponse.state = agentResponse.state;
    connectionResponse.connectionName = agentResponse.theirLabel;
    connectionResponse.alias = agentResponse.alias;
    connectionResponse.did = agentResponse.did;
    connectionResponse.invitationDid = agentResponse.invitationDid;
    connectionResponse.outOfBandId = agentResponse.outOfBandId;
    connectionResponse.createdAt = agentResponse.createdAt;
    connectionResponse.updatedAt = agentResponse.updatedAt;
    connectionResponse.imageUrl = agentResponse.imageUrl;

    return connectionResponse;
  };

  deleteConnectionById = async (id: string): Promise<void> => {
    await this.askar.agent.connections.deleteById(id);
  };

  fetchSchemas = async (): Promise<SchemaRecordDto[]> => {
    let schemaRecords =
      await this.askar.agent.modules.anoncreds.getCreatedSchemas({});
    schemaRecords = schemaRecords.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const schemaResponses = schemaRecords.map((singleSchemaRes) => {
      const schemaResponse = new SchemaRecordDto();
      schemaResponse.id = singleSchemaRes.schemaId;
      schemaResponse.createdAt = singleSchemaRes.createdAt;
      schemaResponse.updatedAt = singleSchemaRes.updatedAt;
      schemaResponse.name = singleSchemaRes.schema.name;
      schemaResponse.attributes = singleSchemaRes.schema.attrNames;
      schemaResponse.version = singleSchemaRes.schema.version;
      schemaResponse.issuerId = singleSchemaRes.schema.issuerId;
      schemaResponse.methodName = singleSchemaRes.methodName;

      return schemaResponse;
    });

    return schemaResponses;
  };

  getSchemaById = async (schemaId: string): Promise<SchemaRecordDto> => {
    const agentResponse = await this.askar.agent.modules.anoncreds.getSchema(
      schemaId,
    );

    if (!agentResponse || !agentResponse.schema) {
      throw new EntityNotFoundError();
    }

    const schemaResponse = new SchemaRecordDto();
    schemaResponse.id = agentResponse.schemaId;
    schemaResponse.name = agentResponse.schema.name;
    schemaResponse.attributes = agentResponse.schema.attrNames;
    schemaResponse.version = agentResponse.schema.version;
    schemaResponse.issuerId = agentResponse.schema.issuerId;

    return schemaResponse;
  };

  createSchema = async (
    schema: CreateSchemaRequestDto,
  ): Promise<SchemaRecordDto> => {
    const dids = await this.askar.agent.dids.getCreatedDids({ method: "indy" });

    const schemaResult =
      await this.askar.agent.modules.anoncreds.registerSchema({
        schema: {
          name: schema.name,
          issuerId: dids[0].did,
          attrNames: schema.attributes,
          version: schema.version,
        },
        options: {},
      });

    if (schemaResult.schemaState.state !== "finished") {
      throw new SchemaNotCreatedError();
    }

    const response = new SchemaRecordDto();

    response.name = schemaResult.schemaState.schema.name;
    response.id = schemaResult.schemaState.schemaId;
    response.issuerId = schemaResult.schemaState.schema.issuerId;
    response.version = schemaResult.schemaState.schema.version;
    response.attributes = schemaResult.schemaState.schema.attrNames;

    return response;
  };

  fetchCredentialDefinitions = async (): Promise<CreddefRecordDto[]> => {
    let credentialDefinitions =
      await this.askar.agent.modules.anoncreds.getCreatedCredentialDefinitions(
        {},
      );
    credentialDefinitions = credentialDefinitions.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const response: Array<CreddefRecordDto> = [];
    for (const credDef of credentialDefinitions) {
      const cd = new CreddefRecordDto();
      cd.id = credDef.credentialDefinitionId;
      cd.createdAt = credDef.createdAt;
      cd.updatedAt = credDef.updatedAt;
      cd.schemaId = credDef.credentialDefinition.schemaId;
      cd.issuerId = credDef.credentialDefinition.issuerId;
      cd.tag = credDef.credentialDefinition.tag;
      response.push(cd);
    }

    return response;
  };

  getCredentialDefinitionById = async (
    credentialDefinitionId: string,
  ): Promise<CreddefRecordDto> => {
    const credDefs =
      await this.askar.agent.modules.anoncreds.getCreatedCredentialDefinitions({
        credentialDefinitionId,
      });

    const credDef = credDefs[0] || null;
    if (!credDef) {
      throw new EntityNotFoundError();
    }

    const cd = new CreddefRecordDto();
    cd.id = credDef.credentialDefinitionId;
    cd.createdAt = credDef.createdAt;
    cd.updatedAt = credDef.updatedAt;
    cd.schemaId = credDef.credentialDefinition.schemaId;
    cd.issuerId = credDef.credentialDefinition.issuerId;
    cd.tag = credDef.credentialDefinition.tag;

    return cd;
  };

  createCredentialDefinition = async (
    credentialDefinitionDto: CreateCredentialDefinitionRequestDto,
  ): Promise<CreddefRecordDto> => {
    const dids = await this.askar.agent.dids.getCreatedDids({ method: "indy" });

    const credDef =
      await this.askar.agent.modules.anoncreds.registerCredentialDefinition({
        credentialDefinition: {
          tag: credentialDefinitionDto.tag,
          issuerId: dids[0].did,
          schemaId: credentialDefinitionDto.schemaId,
        },
        options: {
          supportRevocation: credentialDefinitionDto.supportRevocation ?? false,
        },
      });

    if (credDef.credentialDefinitionState.state !== "finished") {
      throw new CredentialNotCreatedError();
    }

    const response = new CreddefRecordDto();
    response.id = credDef.credentialDefinitionState.credentialDefinitionId;
    response.schemaId =
      credDef.credentialDefinitionState.credentialDefinition.schemaId;
    response.issuerId =
      credDef.credentialDefinitionState.credentialDefinition.issuerId;
    response.tag = credDef.credentialDefinitionState.credentialDefinition.tag;

    return response;
  };

  signJsonLdCredential = async (
    credToSign: W3cCredentialDto,
  ): Promise<W3cJsonLdVerifiableCredentialDto> => {
    this.logger.log("Sign json ld credentials");

    const didRecord = await this.getFirstDidWebRecord();
    const verificationMethodList =
      didRecord.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);

    credToSign.id = didRecord.did + "?uuid=" + uuid();
    if (credToSign.credentialSubject) {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      credToSign.credentialSubject.id = credToSign.id;
    }
    credToSign.issuer = didRecord.did;
    credToSign.issuanceDate = new Date().toISOString();
    let credential: W3cCredential;
    try {
      credential = JsonTransformer.fromJSON(credToSign, W3cCredential);
    } catch (e) {
      this.logger.log("Incorrect request parameter", e);
      throw new OcmError(
        "Invalid JSON-LD data format. Please ensure that your JSON-LD contains the following properties: @context, id, type, issuer, issuanceDate, expirationDate, and credentialSubject.",
      );
    }

    const vc = await w3cServ.signCredential(this.askar.agent.context, {
      format: ClaimFormat.LdpVc,
      credential,
      proofType: "Ed25519Signature2018",
      verificationMethod: verificationMethod.id,
    });

    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVC = vc.toJson() as W3cJsonLdVerifiableCredentialDto;
    this.logger.debug(JSON.stringify(jsonVC, null, 2));

    return jsonVC;
  };

  prepareVerifiablePresentationByJsonLdCredId = async (
    credentialRecordId: string,
  ): Promise<W3cJsonLdVerifiablePresentationDto> => {
    const didRecord = await this.getFirstDidWebRecord();
    const verificationMethodList =
      didRecord.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const credFormatData = await this.askar.agent.credentials.getFormatData(
      credentialRecordId,
    );
    if (!credFormatData.credential?.jsonld) {
      throw new OcmError(
        "The JSON-LD credential is either not in your wallet, pending approval, or not in the JSON-LD format.",
      );
    }

    const jsonLd = credFormatData.credential.jsonld;
    const vc = JsonTransformer.fromJSON(jsonLd, W3cJsonLdVerifiableCredential);

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);
    const presentation = await w3cServ.createPresentation({
      credentials: [vc],
      id: didRecord.did + "?uuid=" + uuid(),
    });

    const vp = await w3cServ.signPresentation(this.askar.agent.context, {
      format: ClaimFormat.LdpVp,
      presentation,
      proofPurpose: null,
      proofType: "Ed25519Signature2018",
      challenge: uuid(),
      verificationMethod: verificationMethod.id,
    });

    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVP = vp.toJson() as W3cJsonLdVerifiablePresentationDto;
    this.logger.debug(JSON.stringify(jsonVP, null, 2));

    return jsonVP;
  };

  offerJsonLdCredential = async (
    connectionId: string | undefined,
    credToSign: W3cCredentialDto,
  ): Promise<CredentialOfferResponseDto> => {
    this.logger.log("offerJsonLdCredential", connectionId);

    const didRecord = await this.getFirstDidWebRecord();
    const verificationMethodList =
      didRecord.didDocument?.verificationMethod || [];
    if (!verificationMethodList.length) {
      throw new EntityNotFoundError(
        "DidDocument does not exists or contains no verification methods",
      );
    }

    const verificationMethod = verificationMethodList[0];

    const w3cServ =
      this.askar.agent.context.dependencyManager.resolve(W3cCredentialService);

    credToSign.id = didRecord.did + "?uuid=" + uuid();
    if (credToSign.credentialSubject) {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      credToSign.credentialSubject.id = credToSign.id;
    }
    credToSign.issuer = didRecord.did;
    credToSign.issuanceDate = new Date().toISOString();
    let credential: W3cCredential;
    try {
      credential = JsonTransformer.fromJSON(credToSign, W3cCredential);
    } catch (e) {
      this.logger.log("Incorrect request parameter", e);
      throw new OcmError(
        "Invalid JSON-LD data format. Please ensure that your JSON-LD contains the following properties: @context, id, type, issuer, issuanceDate, expirationDate, and credentialSubject.",
      );
    }

    const vc = await w3cServ.signCredential(this.askar.agent.context, {
      format: ClaimFormat.LdpVc,
      credential,
      proofType: "Ed25519Signature2018",
      verificationMethod: verificationMethod.id,
    });
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    const jsonVC = vc.toJson() as JsonCredential;

    if (!connectionId) {
      // create connection less credential
      const { credentialRecord, message } =
        await this.askar.agent.credentials.createOffer({
          protocolVersion: "v2",
          credentialFormats: {
            jsonld: {
              credential: jsonVC,
              options: {
                proofType: "Ed25519Signature2018",
                proofPurpose: "assertionMethod",
              },
            },
          },
          autoAcceptCredential: AutoAcceptCredential.ContentApproved,
        });

      credentialRecord.setTag("xRole", "issuer");
      await this.askar.agent.credentials.update(credentialRecord);

      const outOfBandRecord = await this.askar.agent.oob.createInvitation({
        messages: [message],
        handshake: false,
      });

      const credentialUrl = outOfBandRecord.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });

      const shortCredentialUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBandRecord.outOfBandInvitation.id}`;

      const dto = new CredentialRecordDto();
      dto.id = credentialRecord.id;
      dto.state = credentialRecord.state;
      dto.connectionId = credentialRecord.connectionId;
      dto.attributes = credentialRecord.credentialAttributes;
      dto.createdAt = credentialRecord.createdAt;
      dto.tags = credentialRecord.getTags();

      return {
        credentialUrl: credentialUrl,
        shortCredentialUrl: shortCredentialUrl,
        credentialRecord: dto,
      };
    }

    const credentialExchangeRecord =
      await this.askar.agent.credentials.offerCredential({
        connectionId: connectionId,
        protocolVersion: "v2",
        credentialFormats: {
          jsonld: {
            credential: jsonVC,
            options: {
              proofType: "Ed25519Signature2018",
              proofPurpose: "assertionMethod",
            },
          },
        },
        autoAcceptCredential: AutoAcceptCredential.ContentApproved,
      });

    this.logger.log(credentialExchangeRecord);

    credentialExchangeRecord.setTag("xRole", "issuer");
    await this.askar.agent.credentials.update(credentialExchangeRecord);

    const dto = new CredentialRecordDto();
    dto.id = credentialExchangeRecord.id;
    dto.state = credentialExchangeRecord.state;
    dto.connectionId = credentialExchangeRecord.connectionId;
    dto.attributes = credentialExchangeRecord.credentialAttributes;
    dto.createdAt = credentialExchangeRecord.createdAt;
    dto.tags = credentialExchangeRecord.getTags();

    return {
      credentialUrl: null,
      shortCredentialUrl: null,
      credentialRecord: dto,
    };
  };

  offerCredential = async (
    offerCredentialDto: OfferCredentialRequestDto,
  ): Promise<CredentialOfferResponseDto> => {
    this.logger.log(
      "Incoming request",
      JSON.stringify(offerCredentialDto, null, 2),
    );

    if (!offerCredentialDto.connectionId) {
      const { credentialRecord, message } =
        await this.askar.agent.credentials.createOffer({
          protocolVersion: "v2",
          credentialFormats: {
            anoncreds: {
              credentialDefinitionId: offerCredentialDto.credentialDefinitionId,
              attributes: offerCredentialDto.attributes,
            },
          },
          autoAcceptCredential: AutoAcceptCredential.ContentApproved,
        });

      credentialRecord.setTag("xRole", "issuer");
      await this.askar.agent.credentials.update(credentialRecord);

      const outOfBandRecord = await this.askar.agent.oob.createInvitation({
        messages: [message],
        handshake: false,
      });

      const credentialUrl = outOfBandRecord.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });

      const shortCredentialUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBandRecord.outOfBandInvitation.id}`;

      const dto = new CredentialRecordDto();
      dto.id = credentialRecord.id;
      dto.state = credentialRecord.state;
      dto.connectionId = credentialRecord.connectionId;
      dto.attributes = credentialRecord.credentialAttributes;
      dto.createdAt = credentialRecord.createdAt;
      dto.tags = credentialRecord.getTags();

      return {
        credentialUrl: credentialUrl,
        shortCredentialUrl: shortCredentialUrl,
        credentialRecord: dto,
      };
    }
    const credentialExchangeRecord =
      await this.askar.agent.credentials.offerCredential({
        protocolVersion: "v2",
        connectionId: offerCredentialDto.connectionId,
        credentialFormats: {
          anoncreds: {
            credentialDefinitionId: offerCredentialDto.credentialDefinitionId,
            attributes: offerCredentialDto.attributes,
          },
        },
      });

    credentialExchangeRecord.setTag("xRole", "issuer");
    await this.askar.agent.credentials.update(credentialExchangeRecord);

    const dto = new CredentialRecordDto();
    dto.id = credentialExchangeRecord.id;
    dto.state = credentialExchangeRecord.state;
    dto.connectionId = credentialExchangeRecord.connectionId;
    dto.attributes = credentialExchangeRecord.credentialAttributes;
    dto.createdAt = credentialExchangeRecord.createdAt;
    dto.tags = credentialExchangeRecord.getTags();

    return {
      credentialUrl: null,
      shortCredentialUrl: null,
      credentialRecord: dto,
    };
  };

  acceptCredential = async (
    acceptCredentialDto: AcceptCredentialDto,
  ): Promise<CredentialRecordDto> => {
    if (acceptCredentialDto.credentialUrl) {
      return this.acceptOobCredentials(acceptCredentialDto.credentialUrl);
    }
    return this.acceptConnectionCredential(acceptCredentialDto.credentialId);
  };

  acceptOobCredentials = async (url: string): Promise<CredentialRecordDto> => {
    // omit await in order to catch received record in the next line
    setTimeout(() => {
      this.askar.agent.oob.receiveInvitationFromUrl(url, {
        autoAcceptConnection: false,
        autoAcceptInvitation: true,
        // reuseConnection: true,
      });
    }, 20);

    const record = await waitForCredentialExchangeRecordSubject(
      this.askar.agentB,
      {
        state: CredentialState.OfferReceived,
      },
    );

    const acceptedRecord = await this.askar.agent.credentials.acceptOffer({
      credentialRecordId: record.id,
    });

    const response = new CredentialRecordDto();
    response.id = acceptedRecord.id;
    response.state = acceptedRecord.state;
    response.connectionId = acceptedRecord.connectionId;
    response.attributes = acceptedRecord.credentialAttributes;
    response.createdAt = acceptedRecord.createdAt;
    response.tags = acceptedRecord.getTags();

    return response;
  };

  acceptConnectionCredential = async (
    credentialRecordId: string,
  ): Promise<CredentialRecordDto> => {
    const credentialExchangeRecord =
      await this.askar.agent.credentials.acceptOffer({
        credentialRecordId,
      });

    const response = new CredentialRecordDto();
    response.id = credentialExchangeRecord.id;
    response.state = credentialExchangeRecord.state;
    response.connectionId = credentialExchangeRecord.connectionId;
    response.attributes = credentialExchangeRecord.credentialAttributes;
    response.createdAt = credentialExchangeRecord.createdAt;
    response.tags = credentialExchangeRecord.getTags();

    return response;
  };

  declineCredential = async (
    credentialRecordId: string,
  ): Promise<CredentialRecordDto> => {
    const credentialExchangeRecord =
      await this.askar.agent.credentials.declineOffer(credentialRecordId);

    // send request to the issuer that the request is declined, to mark it as Abondoned
    await this.askar.agent.credentials.sendProblemReport({
      credentialRecordId: credentialRecordId,
      description: "Decline offer",
    });

    const response = new CredentialRecordDto();
    response.id = credentialExchangeRecord.id;
    response.state = credentialExchangeRecord.state;
    response.connectionId = credentialExchangeRecord.connectionId;
    response.attributes = credentialExchangeRecord.credentialAttributes;
    response.createdAt = credentialExchangeRecord.createdAt;
    response.tags = credentialExchangeRecord.getTags();

    return response;
  };

  fetchCredentials = async (
    filter: CredentialFilterDto,
  ): Promise<CredentialRecordDto[]> => {
    const query: Query<CredentialExchangeRecord>[] = [];

    if (filter.states) {
      const stateQuery: Query<CredentialExchangeRecord> = {
        $or: filter.states.map((state) => ({ state })),
      };
      query.push(stateQuery);
    }

    if (filter.connectionId) {
      const connectionQuery: Query<CredentialExchangeRecord> = {
        connectionId: filter.connectionId,
      };
      query.push(connectionQuery);
    }

    let credentials = await this.askar.agent.credentials.findAllByQuery({
      $and: query,
    });
    credentials = credentials.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const response: CredentialRecordDto[] = [];
    for (const offer of credentials) {
      const t = new CredentialRecordDto();
      t.id = offer.id;
      t.state = offer.state;
      t.connectionId = offer.connectionId;
      t.createdAt = offer.createdAt;
      t.attributes = offer.credentialAttributes;
      t.tags = offer.getTags();
      response.push(t);
    }

    return response;
  };

  getCredentialById = async (
    credentialId: string,
  ): Promise<CredentialRecordDto> => {
    const credentialRecord = await this.askar.agent.credentials.findById(
      credentialId,
    );

    if (!credentialRecord) {
      throw new EntityNotFoundError();
    }

    const credential = new CredentialRecordDto();
    credential.id = credentialRecord.id;
    credential.state = credentialRecord.state;
    credential.connectionId = credentialRecord.connectionId;
    credential.createdAt = credentialRecord.createdAt;
    credential.attributes = credentialRecord.credentialAttributes;
    credential.tags = credentialRecord.getTags();

    return credential;
  };

  getCredentialFormatDataById = async (
    credentialId: string,
  ): Promise<CredentialFormatDataDto> => {

    const formatData = await this.askar.agent.credentials.getFormatData(
      credentialId,
    );

    if (!formatData) {
      throw new EntityNotFoundError();
    }

    const dto = new CredentialFormatDataDto();
    dto.proposalAttributes = formatData.proposalAttributes;
    dto.offerAttributes = formatData.offerAttributes;
    dto.anoncredsProposal = formatData.proposal?.anoncreds;
    dto.anoncredsOffer = formatData.offer?.anoncreds;
    dto.anoncredsRequest = formatData.request?.anoncreds;
    dto.anoncredsCredential = formatData.credential?.anoncreds;

    dto.all = formatData;

    return dto;
  };

  requestProof = async (
    requestProofDto: RequestProofDto,
  ): Promise<RequestProofResponseDto> => {
    this.logger.log(JSON.stringify(requestProofDto, null, 2));
    const requestedAttributes: Record<string, AnonCredsRequestedAttribute> = {};

    for (const attr of requestProofDto.attributes) {
      requestedAttributes[uuid()] = {
        name: attr.attributeName,
        restrictions: [
          {
            schema_id: attr.schemaId,
            cred_def_id: attr.credentialDefinitionId,
          },
        ],
      };
    }

    if (!requestProofDto.connectionId) {
      this.logger.log("connection Id not detected, creating oob proof");
      const { proofRecord, message } =
        await this.askar.agent.proofs.createRequest({
          protocolVersion: "v2",
          proofFormats: {
            anoncreds: {
              name: "proof-request",
              version: "1.0",
              requested_attributes: requestedAttributes,
            },
          },
        });

      proofRecord.setTag("xRole", "requester");
      await this.askar.agent.proofs.update(proofRecord);

      const outOfBandRecord = await this.askar.agent.oob.createInvitation({
        messages: [message],
        handshake: false,
      });

      const proofUrl = outOfBandRecord.outOfBandInvitation.toUrl({
        domain: this.askar.agentConfig.agentPeerAddress,
      });
      const shortProofUrl = `${this.askar.agentConfig.agentPeerAddress}/invitations/${outOfBandRecord.outOfBandInvitation.id}`;

      const dto = new ProofRecordDto();
      dto.id = proofRecord.id;
      dto.connectionId = proofRecord.connectionId;
      dto.state = proofRecord.state;
      dto.updatedAt = proofRecord.updatedAt;
      dto.createdAt = proofRecord.createdAt;
      dto.tags = proofRecord.getTags();

      return {
        proofUrl: proofUrl,
        shortProofUrl: shortProofUrl,
        proofRecord: dto,
      };
    }

    this.logger.log(`${requestProofDto.connectionId} detected, issuing proof`);

    const exchangeRecord = await this.askar.agent.proofs.requestProof({
      protocolVersion: "v2",
      connectionId: requestProofDto.connectionId,
      proofFormats: {
        anoncreds: {
          name: "proof-request",
          version: "1.0",
          requested_attributes: requestedAttributes,
        },
      },
    });

    exchangeRecord.setTag("xRole", "requester");
    await this.askar.agent.proofs.update(exchangeRecord);

    const response = new ProofRecordDto();
    response.id = exchangeRecord.id;
    response.connectionId = exchangeRecord.connectionId;
    response.state = exchangeRecord.state;
    response.updatedAt = exchangeRecord.updatedAt;
    response.createdAt = exchangeRecord.createdAt;
    response.tags = exchangeRecord.getTags();

    return {
      proofUrl: null,
      shortProofUrl: null,
      proofRecord: response,
    };
  };

  deleteCredentialById = async (id: string): Promise<void> => {
    await this.askar.agent.credentials.deleteById(id);
  };

  fetchProofs = async (filter: ProofFilterDto): Promise<ProofRecordDto[]> => {
    const query: Query<ProofExchangeRecord>[] = [];

    if (filter.states) {
      const stateQuery: Query<ProofExchangeRecord> = {
        $or: filter.states.map((state) => ({ state })),
      };
      query.push(stateQuery);
    }

    if (filter.connectionId) {
      const connectionQuery: Query<ProofExchangeRecord> = {
        connectionId: filter.connectionId,
      };
      query.push(connectionQuery);
    }

    let proofs = await this.askar.agent.proofs.findAllByQuery({
      $and: query,
    });
    proofs = proofs.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const response: ProofRecordDto[] = [];
    for (const proof of proofs) {
      const t = new ProofRecordDto();
      t.id = proof.id;
      t.connectionId = proof.connectionId;
      t.state = proof.state;
      t.updatedAt = proof.updatedAt;
      t.createdAt = proof.createdAt;
      t.tags = proof.getTags();

      response.push(t);
    }

    return response;
  };

  getProofById = async (proofRecordId: string): Promise<ProofRecordDto> => {
    const proofRecord = await this.askar.agent.proofs.findById(proofRecordId);

    if (!proofRecord) {
      throw new EntityNotFoundError();
    }

    const proofResponse = new ProofRecordDto();

    proofResponse.id = proofRecord.id;
    proofResponse.connectionId = proofRecord.connectionId;
    proofResponse.state = proofRecord.state;
    proofResponse.updatedAt = proofRecord.updatedAt;
    proofResponse.createdAt = proofRecord.createdAt;
    proofResponse.tags = proofRecord.getTags();

    return proofResponse;
  };

  getProofFormatDataById = async (
    proofRecordId: string,
  ): Promise<ProofFormatDataDto> => {
    const formatData = await this.askar.agent.proofs.getFormatData(
      proofRecordId,
    );

    if (!formatData) {
      throw new EntityNotFoundError();
    }

    const dto = new ProofFormatDataDto();
    dto.anoncredsProposal = formatData.proposal?.anoncreds;
    dto.anoncredsRequest = formatData.request?.anoncreds;
    dto.anoncredsPresentation = formatData.presentation?.anoncreds;
    return dto;
  };

  proofAcceptanceWait = async (
    proofRecordId: string,
  ): Promise<ProofFormatDataDto> => {
    const proofRecord = await this.askar.agent.proofs.findById(proofRecordId);

    if (!proofRecord) {
      throw new EntityNotFoundError();
    }

    if (proofRecord.state === ProofState.Done) {
      return this.getProofFormatDataById(proofRecordId);
    }

    await waitForProofExchangeRecordSubject(this.askar.agentB, {
      proofRecordId,
      state: ProofState.Done,
      timeoutMs: 3 * 60 * 1000, // 3 minutes
    });

    return this.getProofFormatDataById(proofRecordId);
  };

  acceptProof = async (
    acceptProofDto: AcceptProofDto,
  ): Promise<ProofRecordDto> => {
    if (acceptProofDto.proofUrl) {
      return this.acceptOobProof(acceptProofDto.proofUrl);
    }
    return this.acceptConnectionProof(acceptProofDto.proofId);
  };

  acceptOobProof = async (url: string): Promise<ProofRecordDto> => {
    // omit await in order to catch received record in the next line
    setTimeout(() => {
      this.askar.agent.oob.receiveInvitationFromUrl(url, {
        autoAcceptConnection: false,
        autoAcceptInvitation: true,
        // reuseConnection: true,
      });
    }, 20);

    const record = await waitForProofExchangeRecordSubject(this.askar.agentB, {
      state: ProofState.RequestReceived,
    });

    const requestedCredentials =
      await this.askar.agent.proofs.selectCredentialsForRequest({
        proofRecordId: record.id,
      });

    const acceptedRecord = await this.askar.agent.proofs.acceptRequest({
      proofRecordId: record.id,
      proofFormats: requestedCredentials.proofFormats,
    });

    const response = new ProofRecordDto();

    response.id = acceptedRecord.id;
    response.connectionId = acceptedRecord.connectionId;
    response.state = acceptedRecord.state;
    response.updatedAt = acceptedRecord.updatedAt;
    response.createdAt = acceptedRecord.createdAt;
    response.tags = acceptedRecord.getTags();

    return response;
  };

  acceptConnectionProof = async (
    proofRecordId: string,
  ): Promise<ProofRecordDto> => {
    this.logger.log(`accepting proof request for ${proofRecordId}`);
    const requestedCredentials =
      await this.askar.agent.proofs.selectCredentialsForRequest({
        proofRecordId,
      });

    this.logger.log(JSON.stringify(requestedCredentials, null, 2));

    const proof = await this.askar.agent.proofs.acceptRequest({
      proofRecordId,
      proofFormats: requestedCredentials.proofFormats,
    });

    this.logger.log(JSON.stringify(proof, null, 2));

    const response = new ProofRecordDto();
    response.id = proof.id;
    response.connectionId = proof.connectionId;
    response.state = proof.state;
    response.updatedAt = proof.updatedAt;
    response.createdAt = proof.createdAt;
    response.tags = proof.getTags();

    return response;
  };

  declineProofRequest = async (
    proofRecordId: string,
  ): Promise<ProofRecordDto> => {
    const resultFromDecline = await this.askar.agent.proofs.declineRequest({
      proofRecordId,
      sendProblemReport: true,
    });

    const declineResponse = new ProofRecordDto();
    declineResponse.id = resultFromDecline.id;
    declineResponse.connectionId = resultFromDecline.connectionId;
    declineResponse.state = resultFromDecline.state;
    declineResponse.updatedAt = resultFromDecline.updatedAt;
    declineResponse.createdAt = resultFromDecline.createdAt;
    declineResponse.tags = resultFromDecline.getTags();

    return declineResponse;
  };

  deleteProofById = async (id: string): Promise<void> => {
    await this.askar.agent.proofs.deleteById(id);
  };

  resolve = async (did: string) => {
    return this.askar.agent.dids.resolve(did);
  };

  fetchBasicMessages = async (
    filter: MessageFilterDto,
  ): Promise<MessageRecordDto[]> => {
    const query: Query<BasicMessageRecord>[] = [];

    if (filter.role) {
      const roleQuery: Query<BasicMessageRecord> = {
        role: filter.role,
      };
      query.push(roleQuery);
    }

    if (filter.connectionId) {
      const connectionQuery: Query<BasicMessageRecord> = {
        connectionId: filter.connectionId,
      };
      query.push(connectionQuery);
    }

    let messages = await this.askar.agent.basicMessages.findAllByQuery({
      $and: query,
    });
    messages = messages.sort(
      (a, b) =>
        new Date(b.createdAt).getTime() - new Date(a.createdAt).getTime(),
    );

    const connectionIds = messages.map((message) => message.connectionId);

    const connections = await this.askar.agent.connections.findAllByQuery({
      $or: connectionIds.map((p) => ({ id: p })),
    });
    const grouppedConnections = connections.reduce(
      (acc: { [connId: string]: ConnectionRecord }, conn) => {
        acc[conn.id] = conn;
        return acc;
      },
      {},
    );

    const response: MessageRecordDto[] = [];
    for (const message of messages) {
      const connection = grouppedConnections[message.connectionId];
      const label = connection?.theirLabel || "";

      const t = new MessageRecordDto();
      t.id = message.id;
      t.createdAt = message.createdAt;
      t.updatedAt = message.updatedAt;
      t.connectionId = message.connectionId;
      t.role = message.role;
      t.content = message.content;
      t.sentTime = message.sentTime;
      t.from = message.role === BasicMessageRole.Receiver ? label : "";
      t.to = message.role === BasicMessageRole.Sender ? label : "";

      response.push(t);
    }

    return response;
  };

  sendMessage = async (
    dto: MakeBasicMessageRequestDto,
  ): Promise<MessageRecordDto> => {
    const messageRecord = await this.askar.agent.basicMessages.sendMessage(
      dto.connectionId,
      dto.message,
    );

    const connRecord = await this.askar.agent.connections.findById(
      dto.connectionId,
    );

    const response = new MessageRecordDto();
    response.id = messageRecord.id;
    response.createdAt = messageRecord.createdAt;
    response.updatedAt = messageRecord.updatedAt;
    response.connectionId = messageRecord.connectionId;
    response.role = messageRecord.role;
    response.content = messageRecord.content;
    response.sentTime = messageRecord.sentTime;
    response.from = "";
    response.to = connRecord?.theirLabel || "";

    return response;
  };

  deleteMessageById = async (id: string): Promise<void> => {
    await this.askar.agent.basicMessages.deleteById(id);
  };

  getCreatedDids = async (): Promise<DidRecordDto[]> => {
    const didRecords = await this.askar.agent.dids.getCreatedDids();
    return didRecords.map((p) => {
      const dto = new DidRecordDto();
      const tags = p.getTags();

      dto.did = p.did;
      dto.role = p.role;
      dto.method = tags.method;
      dto.tags = tags;

      return dto;
    });
  };

  trustPingToConnection = async (connectionId: string) => {
    const msg = await this.askar.agent.connections.sendPing(connectionId, {
      responseRequested: true,
      withReturnRouting: false,
    });

    const response = new BaseRecordDto();
    response.id = msg.toJSON()["@id"];

    return response;
  };

  private getFirstDidWebRecord = async (): Promise<DidRecord> => {
    const didWebs = await this.askar.agent.dids.getCreatedDids({
      method: "web",
    });
    if (!didWebs.length) {
      throw new EntityNotFoundError("Agent does not have did:web");
    }
    return didWebs[0];
  };
}
